@@ -1,8 +1,11 @@
 #!/usr/bin/env python3
+# -*- coding: utf-8 -*-
 
 import numpy as np
 import boost_histogram as bh
 import matplotlib.pyplot as plt
+import functools
+import operator
 
 # Make a 2D histogram
 hist = bh.Histogram(bh.axis.Regular(50, -3, 3), bh.axis.Regular(50, -3, 3))
@@ -11,7 +14,7 @@ hist = bh.Histogram(bh.axis.Regular(50, -3, 3), bh.axis.Regular(50, -3, 3))
 hist.fill(np.random.normal(size=1_000_000), np.random.normal(size=1_000_000))
 
 # Compute the areas of each bin
-areas = np.prod(hist.axes.widths, axis=0)
+areas = functools.reduce(operator.mul, hist.axes.widths)
 
 # Compute the density
 density = hist.view() / hist.sum() / areas
