@@ -1,3 +1,4 @@
+# -*- coding: utf-8 -*-
 from __future__ import absolute_import, division, print_function
 
 from .kwargs import KWArgs
@@ -8,7 +9,8 @@ from .axis import Axis
 from .axistuple import AxesTuple
 from .sig_tools import inject_signature
 from .storage import Double, Storage
-from .utils import cast, register, set_family, MAIN_FAMILY, CPP_FAMILY, set_module
+from .utils import cast, register, set_family, MAIN_FAMILY, set_module
+from .six import string_types
 
 import warnings
 import copy
@@ -28,6 +30,16 @@ _histograms = (
 )
 
 
+def _fill_cast(value):
+    """
+    Convert to NumPy arrays. Some buffer objects do not get converted by forcecast.
+    """
+    if isinstance(value, string_types + (bytes,)) or not hasattr(value, "__iter__"):
+        return value
+    else:
+        return np.ascontiguousarray(value)
+
+
 def _hist_or_val(other):
     return other._hist if hasattr(other, "_hist") else other
 
@@ -69,8 +81,14 @@ def _expand_ellipsis(indexes, rank):
 # We currently do not cast *to* a histogram, but this is consistent
 # and could be used later.
 @register(_histograms)
-class BaseHistogram(object):
-    @inject_signature("self, *axes, storage=Double()", locals={"Double": Double})
+@set_family(MAIN_FAMILY)
+@set_module("boost_histogram")
+class Histogram(object):
+    __slots__ = ("_hist", "axes", "metadata")
+
+    @inject_signature(
+        "self, *axes, storage=Double(), metadata=None", locals={"Double": Double}
+    )
     def __init__(self, *axes, **kwargs):
         """
         Construct a new histogram.
@@ -85,20 +103,27 @@ class BaseHistogram(object):
             Provide 1 or more axis instances.
         storage : Storage = bh.storage.Double()
             Select a storage to use in the histogram
+        metadata : Any = None
+            Data that is passed along if a new histogram is created
         """
 
         # Allow construction from a raw histogram object (internal)
         if not kwargs and len(axes) == 1 and isinstance(axes[0], _histograms):
             self._hist = axes[0]
+            self.metadata = None
+            self.axes = self._generate_axes_()
             return
 
-        if not kwargs and len(axes) == 1 and isinstance(axes[0], BaseHistogram):
-            self._hist = copy.copy(axes[0]._hist)
+        # If we construct with another Histogram, support that too
+        if not kwargs and len(axes) == 1 and isinstance(axes[0], Histogram):
+            self.__init__(axes[0]._hist)
+            self.metadata = axes[0].metadata
             return
 
         # Keyword only trick (change when Python2 is dropped)
         with KWArgs(kwargs) as k:
             storage = k.optional("storage", Double())
+            self.metadata = k.optional("metadata")
 
         # Check for missed parenthesis or incorrect types
         if not isinstance(storage, Storage):
@@ -110,7 +135,7 @@ class BaseHistogram(object):
                 raise KeyError("Only storages allowed in storage argument")
 
         # Allow a tuple to represent a regular axis
-        axes = [_arg_shortcut(arg) for arg in axes]
+        axes = tuple(_arg_shortcut(arg) for arg in axes)
 
         if len(axes) > _core.hist._axes_limit:
             raise IndexError(
@@ -121,19 +146,64 @@ class BaseHistogram(object):
         for h in _histograms:
             if isinstance(storage, h._storage_type):
                 self._hist = h(axes, storage)
+                self.axes = self._generate_axes_()
                 return
 
         raise TypeError("Unsupported storage")
 
+    def _generate_axes_(self):
+        """
+        This is called to fill in the axes. Subclasses can override it if they need
+        to change the axes tuple.
+        """
+
+        return AxesTuple(self._axis(i) for i in range(self.ndim))
+
+    def _new_hist(self, _hist):
+        """
+        Return a new histogram given a new _hist, copying metadata.
+        """
+
+        other = self.__class__(_hist)
+        other.metadata = self.metadata
+        return other
+
+    @property
+    def ndim(self):
+        """
+        Number of axes (dimensions) of histogram.
+        """
+        return self._hist.rank()
+
+    def view(self, flow=False):
+        """
+        Return a view into the data, optionally with overflow turned on.
+        """
+        return _to_view(self._hist.view(flow))
+
     def __array__(self):
-        return _to_view(self._hist.view(False))
+        return self.view(False)
 
     def __add__(self, other):
-        return self.__class__(self._hist.__add__(other._hist))
+        if hasattr(other, "_hist"):
+            return self._new_hist(self._hist.__add__(other._hist))
+        else:
+            retval = self.copy()
+            retval += other
+            return retval
 
     def __iadd__(self, other):
-        self._hist.__iadd__(other._hist)
-        return self
+        if isinstance(other, (int, float)) and other == 0:
+            return self
+        if hasattr(other, "_hist"):
+            self._hist.__iadd__(other._hist)
+            # Addition may change category axes
+            self.axes = self._generate_axes_()
+            return self
+        return NotImplemented
+
+    def __radd__(self, other):
+        return self + other
 
     def __eq__(self, other):
         return self._hist == other._hist
@@ -143,7 +213,7 @@ class BaseHistogram(object):
 
     # If these fail, the underlying object throws the correct error
     def __mul__(self, other):
-        return self.__class__(self._hist.__mul__(other))
+        return self._new_hist(self._hist.__mul__(other))
 
     def __rmul__(self, other):
         return self * other
@@ -153,26 +223,40 @@ class BaseHistogram(object):
         return self
 
     def __truediv__(self, other):
-        return self.__class__(self._hist.__truediv__(_hist_or_val(other)))
+        if isinstance(other, Histogram):
+            result = self.copy()
+            result.__itruediv__(other)
+            return result
+        else:
+            return self._new_hist(self._hist.__truediv__(_hist_or_val(other)))
 
     def __div__(self, other):
-        return self.__class__(self._hist.__div__(_hist_or_val(other)))
+        if isinstance(other, Histogram):
+            result = self.copy()
+            result.__idiv__(other)
+            return result
+        else:
+            return self._new_hist(self._hist.__div__(_hist_or_val(other)))
 
     def __itruediv__(self, other):
-        self._hist.__itruediv__(_hist_or_val(other))
+        if isinstance(other, Histogram):
+            view = self.view(flow=True)
+            view.__itruediv__(other.view(flow=True))
+        else:
+            self._hist.__itruediv__(_hist_or_val(other))
         return self
 
     def __idiv__(self, other):
-        self._hist.__idiv__(_hist_or_val(other))
+        if isinstance(other, Histogram):
+            view = self.view(flow=True)
+            view.__idiv__(other.view(flow=True))
+        else:
+            self._hist.__idiv__(_hist_or_val(other))
         return self
 
-    def __copy__(self):
-        other = self.__class__.__new__(self.__class__)
-        other._hist = copy.copy(self._hist)
-        return other
-
+    # TODO: Marked as too complex by flake8. Should be factored out a bit.
     @inject_signature("self, *args, weight=None, sample=None, threads=None")
-    def fill(self, *args, **kwargs):
+    def fill(self, *args, **kwargs):  # noqa: C901
         """
         Insert data into the histogram.
 
@@ -190,15 +274,17 @@ class BaseHistogram(object):
             available threads (usually two per core).
         """
 
+        # Convert to NumPy arrays
+        args = (_fill_cast(a) for a in args)
+
         threads = kwargs.pop("threads", None)
 
-        if threads is None or threads is 1:
+        if threads is None or threads == 1:
             self._hist.fill(*args, **kwargs)
         else:
             if threads == 0:
                 threads = os.cpu_count()
 
-            print(self._hist._storage_type)
             if (
                 self._hist._storage_type is _core.storage.mean
                 or self._hist._storage_type is _core.storage.weighted_mean
@@ -253,22 +339,11 @@ class BaseHistogram(object):
 
         return self
 
-    def __repr__(self):
-        ret = "{self.__class__.__name__}(\n  ".format(self=self)
-        ret += ",\n  ".join(repr(self._axis(i)) for i in range(self._hist.rank()))
-        ret += ",\n  storage={0}".format(self._storage_type())
-        ret += ")"
-        outer = self._hist.sum(flow=True)
-        if outer:
-            inner = self._hist.sum(flow=False)
-            ret += " # Sum: {0}".format(inner)
-            if inner != outer:
-                ret += " ({0} with flow)".format(outer)
-        return ret
-
     def __str__(self):
         """
-        A rendering of the histogram is made using ASCII or unicode characters (whatever is supported by the terminal). What exactly is displayed is still experimental. Do not rely on any particular rendering.
+        A rendering of the histogram is made using ASCII or unicode characters
+        (whatever is supported by the terminal). What exactly is displayed is
+        still experimental. Do not rely on any particular rendering.
         """
         # TODO check the terminal width and adjust the presentation
         # only use for 1D, fall back to repr for ND
@@ -280,7 +355,7 @@ class BaseHistogram(object):
             s = repr(self)
         return s
 
-    def _axis(self, i):
+    def _axis(self, i=0):
         """
         Get N-th axis.
         """
@@ -291,86 +366,27 @@ class BaseHistogram(object):
         return cast(self, self._hist._storage_type, Storage)
 
     def _reduce(self, *args):
-        return self.__class__(self._hist.reduce(*args))
-
-
-# C++ version of histogram
-@set_family(CPP_FAMILY)
-@set_module("boost_histogram.cpp")
-class histogram(BaseHistogram):
-    axis = BaseHistogram._axis
-
-    def rank(self):
-        """
-        Number of axes (dimensions) of histogram.
-        """
-        return self._hist.rank()
-
-    def size(self):
-        """
-        Total number of bins in the histogram (including underflow/overflow).
-        """
-        return self._hist.size()
-
-    def at(self, *indexes):
-        """
-        Select a contents given indices. -1 is the underflow bin, N is the overflow bin.
-        """
-        return self._hist.at(*indexes)
-
-    # Call uses fill since it supports strings,
-    # runtime argument list, etc.
-    @inject_signature("self, *args, weight=None, sample=None")
-    def __call__(self, *args, **kwargs):
-        args = (((a,) if isinstance(a, str) else a) for a in args)
-        self._hist.fill(*args, **kwargs)
-        return self
-
-    def _reset(self):
-        self._hist.reset()
-        return self
-
-    def _empty(self, flow=False):
-        return self._hist.empty(flow)
-
-    def _sum(self, flow=False):
-        return self._hist.sum(flow)
-
-    def _project(self, *args):
-        return self.__class__(self._hist.project(*args))
-
-
-@set_family(MAIN_FAMILY)
-@set_module("boost_histogram")
-class Histogram(BaseHistogram):
-    @inject_signature("self, *axes, storage=Double()", locals={"Double": Double})
-    def __init__(self, *args, **kwargs):
-        super(Histogram, self).__init__(*args, **kwargs)
-
-        # If this is a property, tab completion in IPython does not work
-        self.axes = AxesTuple(self._axis(i) for i in range(self.rank))
-
-    __init__.__doc__ = BaseHistogram.__init__.__doc__
+        return self._new_hist(self._hist.reduce(*args))
 
     def __copy__(self):
-        other = super(Histogram, self).__copy__()
-        other.axes = AxesTuple(other._axis(i) for i in range(other.rank))
+        other = self._new_hist(copy.copy(self._hist))
         return other
 
     def __deepcopy__(self, memo):
         other = self.__class__.__new__(self.__class__)
         other._hist = copy.deepcopy(self._hist, memo)
-        other.axes = AxesTuple(other._axis(i) for i in range(other.rank))
+        other.metadata = copy.deepcopy(self.metadata, memo)
+        other.axes = other._generate_axes_()
         return other
 
     def __getstate__(self):
-        state = self.__dict__.copy()
-        del state["axes"]  # Don't save the cashe
+        state = {"_hist": self._hist, "metadata": self.metadata}
         return state
 
     def __setstate__(self, state):
-        self.__dict__.update(state)
-        self.axes = AxesTuple(self._axis(i) for i in range(self.rank))
+        self._hist = state["_hist"]
+        self.metadata = state["metadata"]
+        self.axes = self._generate_axes_()
 
     def __repr__(self):
         newline = "\n  "
@@ -386,7 +402,7 @@ class Histogram(BaseHistogram):
             else " "
             if len(self.axes) > 0
             else "",
-            comma="," if len(self.axes) > 0 else "",
+            comma=sep,
         )
         ret += ")"
         outer = self.sum(flow=True)
@@ -425,7 +441,7 @@ class Histogram(BaseHistogram):
         # Allow [bh.loc(...)] to work
         for i in range(len(indexes)):
             if callable(indexes[i]):
-                indexes[i] = indexes[i](cast(self, hist.axis(i), Axis))
+                indexes[i] = indexes[i](self.axes[i])
             elif hasattr(indexes[i], "flow"):
                 if indexes[i].flow == 1:
                     indexes[i] = hist.axis(i).size
@@ -438,10 +454,19 @@ class Histogram(BaseHistogram):
 
         return indexes
 
-    def to_numpy(self, flow=False):
+    @inject_signature("self, flow=False, *, dd=False")
+    def to_numpy(self, flow=False, **kwargs):
         """
         Convert to a Numpy style tuple of return arrays.
 
+        Parameters
+        ----------
+
+        flow : bool = False
+            Include the flow bins.
+        dd : bool = False
+            Use the histogramdd return syntax, where the edges are in a tuple
+
         Return
         ------
         contents : Array[Any]
@@ -449,7 +474,16 @@ class Histogram(BaseHistogram):
         *edges : Array[float]
             The edges for each dimension
         """
-        return self._hist.to_numpy(flow)
+
+        with KWArgs(kwargs) as kw:
+            dd = kw.optional("dd", False)
+
+        return_tuple = self._hist.to_numpy(flow)
+
+        if dd:
+            return return_tuple[0], return_tuple[1:]
+        else:
+            return return_tuple
 
     @inject_signature("self, *, deep=True")
     def copy(self, **kwargs):
@@ -468,12 +502,6 @@ class Histogram(BaseHistogram):
         else:
             return copy.copy(self)
 
-    def view(self, flow=False):
-        """
-        Return a view into the data, optionally with overflow turned on.
-        """
-        return _to_view(self._hist.view(flow))
-
     def reset(self):
         """
         Reset bin counters to default values.
@@ -497,8 +525,10 @@ class Histogram(BaseHistogram):
     @property
     def rank(self):
         """
-        Number of axes (dimensions) of histogram.
+        Number of axes (dimensions) of histogram. DEPRECATED, use ndim.
         """
+        msg = "Use .ndim instead"
+        warnings.warn(msg, FutureWarning)
         return self._hist.rank()
 
     @property
@@ -515,10 +545,14 @@ class Histogram(BaseHistogram):
         """
         return self.axes.size
 
-    def __getitem__(self, index):
+    # TODO: Marked as too complex by flake8. Should be factored out a bit.
+    def __getitem__(self, index):  # noqa: C901
 
         indexes = self._compute_commonindex(index)
 
+        # Workaround for missing "pick" and no reduction on bool axes
+        pick_bool = {}
+
         # If this is (now) all integers, return the bin contents
         # But don't try *dict!
         if not hasattr(indexes, "items"):
@@ -529,72 +563,63 @@ class Histogram(BaseHistogram):
 
         integrations = set()
         slices = []
-        zeroes_start = []
-        zeroes_stop = []
-
-        # We could use python's sum here, but for now, a private sum is used
-        class ext_sum:
-            projection = True
 
         # Compute needed slices and projections
         for i, ind in enumerate(indexes):
             if hasattr(ind, "__index__"):
-                ind = slice(ind.__index__(), ind.__index__() + 1, ext_sum())
+                if isinstance(self.axes[i]._ax, _core.axis.boolean):
+                    pick_bool[i] = ind
+                    ind = slice(None, None, sum)
+                else:
+                    ind = slice(ind.__index__(), ind.__index__() + 1, sum)
 
             elif not isinstance(ind, slice):
                 raise IndexError(
                     "Must be a slice, an integer, or follow the locator protocol."
                 )
+            start, stop = self.axes[i]._process_loc(ind.start, ind.stop)
+
             if ind != slice(None):
                 merge = 1
                 if ind.step is not None:
-                    if hasattr(ind.step, "projection"):
-                        if ind.step.projection:
+                    if hasattr(ind.step, "factor"):
+                        merge = ind.step.factor
+                    elif callable(ind.step):
+                        if ind.step is sum:
                             integrations.add(i)
-                            if ind.start is not None:  # TODO: Support callables too
-                                zeroes_start.append(i)
-                            if ind.stop is not None:
-                                zeroes_stop.append(i)
-                            if ind.stop is None and ind.start is None:
-                                continue
-                        elif hasattr(ind.step, "factor"):
-                            merge = ind.step.factor
                         else:
-                            raise IndexError("Invalid rebin, must have integer .factor")
+                            raise RuntimeError("Full UHI not supported yet")
+
+                        if ind.start is not None or ind.stop is not None:
+                            slices.append(
+                                _core.algorithm.slice(
+                                    i, start, stop, _core.algorithm.slice_mode.crop
+                                )
+                            )
+                        continue
                     else:
                         raise IndexError(
                             "The third argument to a slice must be rebin or projection"
                         )
 
-                process_loc = (
-                    lambda x, y: y
-                    if x is None
-                    else x(self._axis(i))
-                    if callable(x)
-                    else x
-                )
-                begin = process_loc(ind.start, 0)
-                end = process_loc(ind.stop, len(self._axis(i)))
+                slices.append(_core.algorithm.slice_and_rebin(i, start, stop, merge))
 
-                slices.append(_core.algorithm.slice_and_rebin(i, begin, end, merge))
+        for i, select in pick_bool.items():
+            self = self.copy()
+            view = self.view(flow=True)
+            view[
+                tuple(~select if j == i else slice(None) for j in range(self.ndim))
+            ] = 0
 
-        reduced = self._reduce(*slices)
+        reduced = self._hist.reduce(*slices)
 
         if not integrations:
-            return self.__class__(reduced)
+            return self._new_hist(reduced)
         else:
-            projections = [i for i in range(self.rank) if i not in integrations]
-
-            # Replacement for crop missing in BH
-            for i in zeroes_start:
-                if self.axes[i].options.underflow:
-                    reduced._hist._reset_row(i, -1)
-            for i in zeroes_stop:
-                if self.axes[i].options.overflow:
-                    reduced._hist._reset_row(i, reduced.axes[i].size)
+            projections = [i for i in range(self.ndim) if i not in integrations]
 
             return (
-                self.__class__(reduced.project(*projections))
+                self._new_hist(reduced.project(*projections))
                 if projections
                 else reduced.sum(flow=True)
             )
@@ -624,14 +649,23 @@ class Histogram(BaseHistogram):
         """
         indexes = self._compute_commonindex(index)
 
-        if isinstance(value, BaseHistogram):
+        if isinstance(value, Histogram):
             raise TypeError("Not supported yet")
 
         value = np.asarray(value)
         view = self.view(flow=True)
 
+        # Shortcut: allow raw arrays for WeightedSum
+        weighted_dtype = np.dtype([("value", "<f8"), ("variance", "<f8")])
+        if (
+            view.dtype == weighted_dtype != value.dtype
+            and len(value.dtype) != 2
+            and value.ndim > 0
+            and value.shape[-1] == 2
+        ):
+            value = value.astype(np.double).view(weighted_dtype)[..., 0]
         # Disallow mismatched data types
-        if len(value.dtype) != len(view.dtype):
+        elif len(value.dtype) != len(view.dtype):
             raise ValueError("Mismatched data types; matching types required")
 
         # Numpy does not broadcast partial slices, but we would need
@@ -702,4 +736,4 @@ class Histogram(BaseHistogram):
         those axes only. Flow bins are used if available.
         """
 
-        return self.__class__(self._hist.project(*args))
+        return self._new_hist(self._hist.project(*args))
