@@ -1,5 +1,18 @@
+# -*- coding: utf-8 -*-
 from __future__ import absolute_import, division, print_function
 
+from .._internal.axis import Axis
+from .._internal.axis import (
+    Regular,
+    Variable,
+    Integer,
+    IntCategory,
+    StrCategory,
+    Boolean,
+)
+from .._core.axis import options
+from . import transform
+
 del absolute_import, division, print_function
 
 __all__ = (
@@ -8,12 +21,8 @@ __all__ = (
     "Integer",
     "IntCategory",
     "StrCategory",
+    "Boolean",
     "Axis",
     "options",
     "transform",
 )
-
-from .._internal.axis import Axis, options
-from .._internal.utils import register
-from .._internal.axis import Regular, Variable, Integer, IntCategory, StrCategory
-from . import transform
